-- Section 4.2
-- Expressions with Declarations
-- Following code is an extension of x3-4.g

-- Lexical elements
data Tag = T_Num | T_Ide | T_Sym | T_Junk

instance Eq Tag where
	T_Num==T_Num = True
	T_Ide==T_Ide = True
	T_Sym==T_Sym = True
	T_Junk==T_Junk = True
	_ == _ = False

-- Lexical analysis
lit :: [Char] -> Parser (Token Tag) [Char]
lit xs = literal (T_Sym,xs) `using` snd

ident :: Parser Char [Char]
ident = (satisfy isAlpha `then_` many (satisfy isAlphanum)) `using` cons

lexer :: Parser Char [Token Tag]
lexer = lexime [(some (satisfy isSpace), T_Junk),
		(number, T_Num),
		(string "let", T_Sym),
		(string "in", T_Sym),
		(ident, T_Ide),
		(anyof string ["(",")","+","-","*","/","="], T_Sym) ]

-- Syntactic domains
data Expr = Num Numeral | Var Ide | Bexpr Expr Opr Expr | Let Decl Expr
data Decl = Decl Ide Expr
type Numeral = [Char]
type Ide = [Char]
data Opr = Plus | Minus | Times | Over

-- Syntax analysis
expr =	((lit "let" `xthen` decl `then_` lit "in" `xthen` expr)
		`using` (\(d,e)->Let d e))
	`alt` expr'
expr' =	((term `then_` lit "+" `xthen` term) `using` plus) `alt`
	((term `then_` lit "-" `xthen` term) `using` minus) `alt`
	term
term =	((factor `then_` lit "*" `xthen` factor) `using` times) `alt`
	((factor `then_` lit "/" `xthen` factor) `using` over) `alt`
	factor
factor = (kind T_Num `using` (\s->Num s)) `alt`
	 (kind T_Ide `using` (\s->Var s)) `alt`
	 (lit "(" `xthen` expr `thenx` lit ")")
decl =	kind T_Ide `then_` lit "=" `xthen` expr `using` (\(x,e)->Decl x e)

plus (x,y) = Bexpr x Plus y
minus (x,y) = Bexpr x Minus y
times (x,y) = Bexpr x Times y
over (x,y) = Bexpr x Over y

-- Semantic functions
numval :: Numeral -> Int
numval = foldl1 (*+) . map digval
	 where
	 a *+ b = a * 10 + b

digval :: Char -> Int
digval a = ord a - ord '0'

type Env = Ide -> Int

expval :: Expr -> Env -> Int
expval (Num n) r = numval n
expval (Var x) r = r x
expval (Bexpr e op e') r = oprfun op (expval e r) (expval e' r)
expval (Let d e) r = expval e (declenv d r r)

declenv :: Decl -> Env -> Env -> Env
declenv (Decl x e) r' r = update r x (expval e r')

update :: Env -> Ide -> Int -> Env
update r x v y | x==y = v
	       | otherwise = r y

oprfun :: Opr -> Int -> Int -> Int
oprfun Plus = (+)
oprfun Minus = (-)
oprfun Times = (*)
oprfun Over = (/)

-- Parser and Evaluator
type Prog = Expr
prog = expr

lex :: [Char] -> [Token Tag]
lex = (strip T_Junk) . fst . head . lexer

syn :: [Token Tag] -> Prog
syn = fst . head . prog

parse :: [Char] -> Prog
parse = syn . lex

-- Initial environment
r0 :: Env
r0 x = undefined

eval e = expval (parse e) r0
