-- Section 5.3
-- Fig.5.7: Variable declarations and storage allocation

cmdexec :: Cmd -> Env -> State -> State
cmdexec (Assign x e) r s@(l,t) = (l, update t k (expval e r s))
		where
		V_Loc k = lookup r x
cmdexec (Block ds cs) r s@(l,t) = (l, t')
	where
	(r',s') = foldl (flip vardecl) (r,s) ds
	(_,t') = foldl (flip (flip cmdexec r')) s' cs

vardecl :: Decl -> (Env,State) -> (Env,State)
vardecl (VarDecl x) (r,s) = (update r x (V_Loc l'),s')
			    where
			    (l',s') = alloc s 1
	

expval :: Expr -> Env -> State -> Val
expval (Num n) r s = V_Int (numval n)
expval (Var x) r (_,t) = lookup t k
			 where
			 V_Loc k = lookup r x
expval (Bexpr o e e') r s = V_Int (binopr o v v')
			    where
			    V_Int v = expval e r s
			    V_Int v' = expval e' r s

-- Storage allocation
alloc :: State -> Int -> (Loc,State)
alloc (l,t) n = (l, (l+n,t'))
	where
	t' = foldl (flip flip undefined . update) t [l,l+n-1]

-- Initial store
t0 :: Store
t0 = none

-- Initial state
s0 :: State
s0 = (1, t0)

-- Initial environment
r0 :: Env
r0 = update none "result" (V_Loc 0)

-- Execution function
exec c = lookup (snd s) 0
	 where
	 s = cmdexec (parse c) r0 s0

-- Interactive execution
exec' = interact (show' . exec)
