-- Section 5.3
-- Fig.5.7: Variable declarations and storage allocation

lexer :: Parser Char [Token Tag]
lexer = lexime [(some (satisfy isSpace), T_Junk),
		(number, T_Num),
		(string "begin", T_Sym),
		(string "end", T_Sym),
		(string "var", T_Sym),
		(ident, T_Ide),
		(anyof string ["(",")","+","-","*","/",":=",",",";"], T_Sym) ]

-- Syntactic domains
data Cmd = Assign Ide Expr | Block [Decl] [Cmd]
data Decl = VarDecl Ide
data Expr = Num Numeral | Var Ide | Bexpr BinOpr Expr Expr

-- Syntax analysis
decls = many (decl `seq_x` lit ";") `using` concat
decl = lit "var" `x_seq`
	(decls' `seq` many (lit "," `x_seq` decls') `using` uncurry (:))
decls' = kind T_Ide `using` VarDecl
cmd =	lit "begin" `x_seq` decls `seq` seqcmd `seq_x` lit "end"
		`using` uncurry Block `alt`
	kind T_Ide `seq` lit ":=" `x_seq` expr
		`using` uncurry Assign
seqcmd= cmd `seq` many (lit ";" `x_seq` cmd) `using` uncurry (:)
expr =	term `seq` expr' `using` uncurry (foldl (flip id))
expr' = many ((lit "+" `x_seq` term `using` flip (Bexpr Plus)) `alt`
		(lit "-" `x_seq` term `using` flip (Bexpr Minus)))
term =	factor `seq` term' `using` uncurry (foldl (flip id))
term' = many ((lit "*" `x_seq` factor `using` flip (Bexpr Times)) `alt`
		(lit "/" `x_seq` factor `using` flip (Bexpr Over)))
factor =kind T_Num `using` Num `alt`
	kind T_Ide `using` Var `alt`
	lit "(" `x_seq` expr `seq_x` lit ")"

-- Parser and Evaluator
type Prog = Cmd
prog = cmd

lex :: [Char] -> [Token Tag]
lex = (strip T_Junk) . fst . head . lexer

syn :: [Token Tag] -> Prog
syn = fst . head . prog

parse :: [Char] -> Prog
parse = syn . lex
