-- Section 5.5
-- Expressions with side-effects

cmdexec :: Cmd -> Env -> State -> State
cmdexec (Assign e e') r s =
		(l, update t k v)
		where
		V_Loc k = explval e r s
		(v,(l,t)) = expexec e' r s
cmdexec (Block ds cs) r s@(l,t) = (l, t')
	where
	(r',s') = foldl (flip vardecl) (r,s) ds
	(_,t') = foldl (flip (flip cmdexec r')) s' cs

-- Variable Declaration
vardecl :: Decl -> (Env,State) -> (Env,State)
vardecl (VarDecl x) (r,s) = (update r x (V_Loc l'),s')
			    where
			    (l',s') = alloc s 1

-- Expression (L-value)
explval :: Expr -> Env -> State -> Val
explval (Num n) r s = undefined
explval (Var x) r s = lookup r x
explval (Bexpr o e e') r s = undefined
explval (Rexpr o e e') r s = undefined

-- Storage allocation
alloc :: State -> Int -> (Loc,State)
alloc (l,t) n = (l, (l+n,t'))
	where
	t' = foldl (flip flip undefined . update) t [l,l+n-1]
-- Expression with side-effect operators
expexec :: Expr -> Env -> State -> (Val,State)
expexec (Num n) r s = (V_Int(numval n), s)
expexec e@(Var _) r s@(_,t) = (lookup t k, s)
			      where
			      V_Loc k = explval e r s
--expexec e@(Ael _ _) r s@(_,t) = (lookup t k, s')
--			         where
--				 (V_Loc k,s') = lvalexec e r s
expexec (Bexpr o e e') r s = (V_Int(binopr o v v'), s'')
			     where
			     (V_Int v,s') = expexec e r s
			     (V_Int v',s'') = expexec e' r s'
expexec (Rexpr o e e') r s = (V_Bool(relopr o v v'), s'')
			     where
			     (V_Int v,s') = expexec e r s
			     (V_Int v', s'') = expexec e' r s'
expexec (AssExpr e e') r s = (v, (l, update t k v))
			     where
			     (v,(l,t)) = expexec e' r s
			     V_Loc k = explval e r s
expexec (SeqExpr es) r s@(l,_) = foldl expexec' (undefined,s) es
				 where
				 expexec' (_,s) e = expexec e r s


-- Initial store
t0 :: Store
t0 = none

-- Initial state
s0 :: State
s0 = (1, t0)

-- Initial environment
r0 :: Env
r0 = update none "result" (V_Loc 0)

-- Execution function
exec c = lookup (snd s) 0
	 where
	 s = cmdexec (parse c) r0 s0

-- Interactive execution
exec' = interact (show' . exec)
