-- Section 8.1 (c)
-- Call by constant

cmdexec :: Cmd -> Env -> Cont -> Cont
cmdexec (Assign e e') r p s@(l,t) =
	p (l, update t k (expval e' r s))
	where
	V_Loc k = explval e r s
cmdexec (Block ds cs) r p s@(l,_) =
	foldr (flip cmdexec r') p' cs s'
	where
	(r',s') = foldl (flip vardecl) (r,s) ds
	p' (_,t) = p (l,t)
cmdexec (If e (c,c')) r p s
	| v = cmdexec c r p s
	| otherwise = cmdexec c' r p s
	where
	V_Bool v = expval e r s
cmdexec c@(While e c') r p s
	| v = cmdexec c' r'' p' s
	| otherwise = p s
	where
	V_Bool v = expval e r s
	r' = update r "*BREAK*" (V_Cont p)
	r'' = update r' "*CONTINUE*" (V_Cont p')
	p' = cmdexec c r p
cmdexec Skip r p s = p s
cmdexec (Output e) r p s = expval e r s : p s
cmdexec Break r _ s = p' s
		      where
		      V_Cont p' = lookup r "*BREAK*"
cmdexec Continue r _ s = p' s
			 where
			 V_Cont p' = lookup r "*CONTINUE*"
cmdexec (Call x [e]) r p s = u [expval e r s] p s
			     where
			     V_Proc u = lookup r x

-- Variable Declaration
vardecl :: Decl -> (Env,State) -> (Env,State)
vardecl (VarDecl x) (r,s) = (update r x (V_Loc l'),s')
			    where
			    (l',s') = alloc s 1
vardecl (ArrDecl x n) (r,s) = (update r x (V_Loc l'),s')
			      where
			      (l',s') = alloc s (numval n)
vardecl (ProcDecl x [y] c) (r,s) = (r',s)
				   where
				   r' = update r x (V_Proc u)
				   u [v] = cmdexec c (update r' y v)

-- Expression (R-value)
expval :: Expr -> Env -> State -> Val
expval (Num n) r s = V_Int (numval n)
expval (Var x) r s@(_,t) =
		case lookup r x of
		V_Loc k -> lookup t k
		v -> v
expval (ArrElem x e) r s@(_,t) = lookup t (k + v)
				 where
				 V_Loc k = lookup r x
				 V_Int v = expval e r s
expval (Bexpr o e e') r s = V_Int (binopr o v v')
			    where
			    V_Int v = expval e r s
			    V_Int v' = expval e' r s
expval (Rexpr o e e') r s = V_Bool (relopr o v v')
			    where
			    V_Int v = expval e r s
			    V_Int v' = expval e' r s

-- Expression (L-value)
explval :: Expr -> Env -> State -> Val
explval (Num n) r s = undefined
explval (Var x) r s@(_,t) =
		case lookup r x of
		v@(V_Loc _) -> v
		_ -> undefined
explval (ArrElem x e) r s = V_Loc (k + v)
			    where
			    V_Loc k = lookup r x
			    V_Int v = expval e r s
explval (Bexpr o e e') r s = undefined
explval (Rexpr o e e') r s = undefined

-- Storage allocation
alloc :: State -> Int -> (Loc,State)
alloc (l,t) n = (l, (l+n,t'))
	where
	t' = foldl (flip flip undefined . update) t [l,l+n-1]

-- Initial store
t0 :: Store
t0 = none

-- Initial state
s0 :: State
s0 = (0, t0)

-- Initial environment
r0 :: Env
r0 = none

-- Initial continuation
stop :: Cont
stop s = []

-- Execution function
exec c = cmdexec (parse c) r0 stop s0

-- Interactive execution
exec' = interact (show' . exec)
