-- Section 8.1 (d)
-- Call by closure

cmdexec :: Cmd -> Env -> Cont -> Cont
cmdexec (Assign e e') r p s = explexec e r p' s
	where
	p' (V_Loc k) s = expexec e' r p'' s
			 where
			 p'' v (l,t) = p (l, update t k v)
cmdexec (Block ds cs) r p s@(l,_) =
	foldr (flip cmdexec r') p' cs s'
	where
	(r',s') = foldl (flip vardecl) (r,s) ds
	p' (_,t) = p (l,t)
cmdexec (If e (c,c')) r p s = expexec e r q s
	where
	q (V_Bool True) = cmdexec c r p
	q (V_Bool False) = cmdexec c' r p
cmdexec c@(While e c') r p s = expexec e r q s
	where
	q (V_Bool True) = cmdexec c' r'' p'
	q (V_Bool False) = p
	r' = update r "*BREAK*" (V_Cont p)
	r'' = update r' "*CONTINUE*" (V_Cont p')
	p' = cmdexec c r p
cmdexec Skip r p s = p s
cmdexec (Output e) r p s = expexec e r q s
			   where
			   q v s = v : p s
cmdexec Break r _ s = p' s
		      where
		      V_Cont p' = lookup r "*BREAK*"
cmdexec Continue r _ s = p' s
			 where
			 V_Cont p' = lookup r "*CONTINUE*"
cmdexec (Call x [e]) r p s = u [V_Thunk (expexec e r)] p s 
			     where
			     V_Proc u = lookup r x

-- Variable Declaration
-- Mutual recursive procs and funcs not allowed
vardecl :: Decl -> (Env,State) -> (Env,State)
vardecl (VarDecl x) (r,s) = (update r x (V_Loc l'),s')
			    where
			    (l',s') = alloc s 1
vardecl (ArrDecl x n) (r,s) = (update r x (V_Loc l'),s')
			      where
			      (l',s') = alloc s (numval n)
vardecl (ProcDecl x [y] c) (r,s) = (r',s)
				   where
				   r' = update r x (V_Proc u)
				   u [v] = cmdexec c (update r' y v)

-- Storage allocation
alloc :: State -> Int -> (Loc,State)
alloc (l,t) n = (l, (l+n,t'))
	where
	t' = foldl (flip flip undefined . update) t [l,l+n-1]

-- Expression with side-effect operators
-- Expression continuation (R-value)
expexec :: Expr -> Env -> E_Cont -> Cont
expexec (Num n) r q = q (V_Int(numval n))
expexec (Var x) r q =
		case lookup r x of
		v@(V_Loc k) -> q' v
			where
			q' (V_Loc k) s@(_,t) = q (lookup t k) s
		V_Thunk g -> g q
		v -> q v
expexec e@(ArrElem _ _) r q = explexec e r q'
			      where
			      q' (V_Loc k) s@(_,t) = q (lookup t k) s
expexec (Bexpr o e e') r q = expexec e r q'
			     where
			     q' (V_Int v) = expexec e' r q''
				where
				q'' (V_Int v') = q (V_Int(binopr o v v'))
expexec (Rexpr o e e') r q = expexec e r q'
			     where
			     q' (V_Int v) = expexec e' r q''
				where
				q'' (V_Int v') = q (V_Bool(relopr o v v'))
expexec (AssExpr e e') r q = explexec e r q'
			     where
			     q' (V_Loc k) = expexec e' r q''
				where
				q'' v (l,t) = q v (l, update t k v)
expexec (SeqExpr es) r q =
	foldl1 (flip (.) ((.) const) . (.)) (map (flip expexec r) es) q
expexec (Apply x _) r q = h [] q
			  where
			  V_Func h = lookup r x

-- Expression execution (L-value)
explexec :: Expr -> Env -> E_Cont -> Cont
explexec (Var x) r q = q (V_Loc k)
		       where
		       V_Loc k = lookup r x
explexec (ArrElem x e) r q = expexec e r q'
			     where
			     V_Loc k = lookup r x
			     q' (V_Int v) = q (V_Loc (k+v))
explexec e r q = undefined

-- Initial store
t0 :: Store
t0 = none

-- Initial state
s0 :: State
s0 = (0, t0)

-- Initial environment
r0 :: Env
r0 = none

-- Initial continuation
stop :: Cont
stop s = []

-- Execution function
exec c = cmdexec (parse c) r0 stop s0

-- Interactive execution
exec' = interact (show' . exec)
