-- Section 8.2
-- Unification
--
-- Expression for Unification
-- a for Identifier, b for structure
data U_Expr a b = U_Var a | U_Struct b [U_Expr a b]
u_vars :: Eq a => U_Expr a b -> [a]
u_vars (U_Var x) = [x]
u_vars (U_Struct _ es) = (nub . concat . (map u_vars)) es

-- Abstract type for Substitution
null_Subst :: Subst a b
single_Subst :: Eq a => a -> U_Expr a b -> Subst a b
apply_Subst :: Eq a => Subst a b -> U_Expr a b -> U_Expr a b
comp_Subst :: Eq a => Subst a b -> Subst a b -> Subst a b

-- Implementation with functions
--type Subst a b = a -> U_Expr a b
--null_Subst x = U_Var x
--single_Subst x e x' | x==x' = e
--		    | otherwise = U_Var x'
--apply_Subst p (U_Var x) = p x
--apply_Subst p (U_Struct c es)
--	= U_Struct c (map (apply_Subst p) es)
--comp_Subst p p' = (apply_Subst p) . p'

-- Alternative implementation
type Subst a b = [(a,U_Expr a b)]
null_Subst = []
single_Subst x e = [(x,e)]
apply_Subst p e = foldr subst e p
	where
	subst (x,e) e'@(U_Var x') | x==x' = e
				  | otherwise = e'
	subst q (U_Struct c es) = U_Struct c (map (subst q) es)
comp_Subst p p' = p ++ p'

type Unifier a b = [Subst a b]

unify_failure :: Unifier a b
unify_failure = []

null_unifier = [[]]

unify :: Eq (a,b) => U_Expr a b -> U_Expr a b -> Unifier a b
unify (U_Var x) e'@(U_Var x') | x==x' = [null_Subst]
			      | otherwise = [single_Subst x e']
unify (U_Var x) e' | x `elem` u_vars e' = unify_failure
		   | otherwise = [single_Subst x e']
unify e (U_Var x') | x' `elem` u_vars e = unify_failure
		   | otherwise = [single_Subst x' e]
unify (U_Struct c es) (U_Struct c' es')
	| c==c' = unify_list es es'
	| otherwise = unify_failure

unify_list :: Eq (a,b) => [U_Expr a b] -> [U_Expr a b] -> Unifier a b
unify_list [] [] = [null_Subst]
unify_list [] (_:_) = unify_failure
unify_list (_:_) [] = unify_failure
unify_list (e:es) (e':es') =
	[comp_Subst p' p | p<-unify e e',
	 p'<-unify_list (map (apply_Subst p) es) (map (apply_Subst p) es')]
