-- Chapter 9
-- Generation of Reverse Polish code for expressions and Continuation code
-- for control structures.

-- Stack represented by a list
type Stack = [Val]
type Store = Assoc Ide Val
data Val = V_Int Int | V_Bool Bool

-- State represented by  Stack and Store
type State = (Stack,Store)

-- Instruction
data Instr = PushNum Int | PushVar Ide | PopVar Ide
	   | Bopr BinOpr | Ropr RelOpr
	   | Jump Code | JumpFalse Code | Stop

type Code = [Instr]

-- Code generation
gencmd :: Cmd -> Code -> Code
gencmd (Assign x e) is = genexp e ++ [PopVar x] ++ is
gencmd (Seq cs) is = foldr gencmd is cs
gencmd (If e (c,c')) is =
	genexp e ++ [JumpFalse is'] ++ gencmd c [Jump is] ++ is'
	where
	is' = gencmd c' is
gencmd (While e c) is = is'
	where
	is' = genexp e ++ [JumpFalse is] ++ gencmd c [Jump is'] ++ is
gencmd Skip p = p

genexp :: Expr -> Code
genexp (Num n) = [PushNum (numval n)]
genexp (Var x) = [PushVar x]
genexp (Bexpr o e e') = genexp e ++ genexp e' ++ [Bopr o]
genexp (Rexpr o e e') = genexp e ++ genexp e' ++ [Ropr o]

-- Code generator
compile s = gencmd (parse s) [Stop]
