-- Fig.9.9: Parsing primitives and combinators
-- parselib.gs

type Parser a b = [a] -> [(b,[a])]

succeed :: b -> Parser a b
succeed v xs = [(v,xs)]

fail :: Parser a b
fail xs = []

satisfy :: (a -> Bool) -> Parser a a
satisfy q [] = fail []
satisfy q (x:xs) | q x = succeed x xs
		 | otherwise = fail xs

literal :: Eq a => a -> Parser a a
literal x = satisfy (==x)

infixr 4 `alt`
alt :: Parser a b -> Parser a b -> Parser a b
(p1 `alt` p2) xs = p1 xs ++ p2 xs

infixr 6 `seq`
seq :: Parser a b -> Parser a c -> Parser a (b,c)
(p1 `seq` p2) xs =
    [((v,v'),xs'')|(v,xs')<-p1 xs, (v',xs'')<-p2 xs']

infixl 5 `using`
using :: Parser a b -> (b -> c) -> Parser a c
(p `using` f) xs = [(f v,xs')|(v,xs')<-p xs]

many :: Parser a b -> Parser a [b]
many p = p `seq` many p `using` uncurry (:) `alt` succeed []

some :: Parser a b -> Parser a [b]
some p = p `seq` many p `using` uncurry (:)

infixr 6 `x_seq`, `seq_x`
x_seq :: Parser a b -> Parser a c -> Parser a c
p1 `x_seq` p2 = (p1 `seq` p2) `using` snd
seq_x :: Parser a b -> Parser a c -> Parser a b
p1 `seq_x` p2 = (p1 `seq` p2) `using` fst

anyof :: (a -> Parser b c) -> [a] -> Parser b c
anyof p = foldr (alt . p) fail
